//******************************************************************************
//
// File:    AnalyzeDigrams.java
// Package: ---
// Unit:    Class AnalyzeDigrams
//
// This Java source file is copyright (C) 2016 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.io.OutStream;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;

/**
 * Class AnalyzeDigrams is a main program that analyzes digram frequencies in a
 * series of English text files.
 * <P>
 * Usage: <TT>java AnalyzeDigrams -c <I>freqfile</I> <I>textfile</I>
 * [<I>textfile</I> ...]</TT>
 * <P>
 * With the <TT>"-c"</TT> flag (create), the program reads the given text file
 * or files, computes the frequencies, and stores the frequencies in the given
 * frequency file, overwriting the frequency file.
 * <P>
 * Usage: <TT>java AnalyzeDigrams -u <I>freqfile</I> <I>textfile</I>
 * [<I>textfile</I> ...]</TT>
 * <P>
 * With the <TT>"-u"</TT> flag (update), the program reads the frequencies from
 * the given frequency file, reads the given text file or files, updates the
 * frequencies, and stores the frequencies back in the given frequency file.
 * <P>
 * Usage: <TT>java AnalyzeDigrams -p <I>freqfile</I></TT>
 * <P>
 * With the <TT>"-p"</TT> flag (update), the program reads the frequencies from
 * the given frequency file and prints them. The frequency file is not updated.
 * <P>
 * The frequency file contains one streamed object: a histogram of the digrams
 * (class {@linkplain DigramHist}).
 *
 * @author  Alan Kaminsky
 * @version 25-Apr-2016
 */
public class AnalyzeDigrams
	{

// Prevent construction.

	private AnalyzeDigrams()
		{
		}

// Main program.

	/**
	 * Main program.
	 */
	public static void main
		(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		if (args.length < 2) usage();
		boolean create = false;
		boolean update = false;
		boolean print = false;
		if (args[0].equals ("-c"))
			create = true;
		else if (args[0].equals ("-u"))
			update = true;
		else if (args[0].equals ("-p"))
			print = true;
		else
			usage();
		File freqfile = new File (args[1]);
		if ((create || update) && args.length < 3) usage();

		// Set up histogram.
		DigramHist hist;
		if (create)
			{
			// Create empty histogram.
			hist = new DigramHist();
			}
		else
			{
			// Read histogram from frequency file.
			InStream in = new InStream (new FileInputStream (freqfile));
			hist = (DigramHist) in.readObject();
			in.close();
			}

		if (create || update)
			{
			// Process all text files.
			for (int f = 2; f < args.length; ++ f)
				{
				System.err.printf ("Processing %s ...%n", args[f]);
				InputStream text =
					new BufferedInputStream (new FileInputStream (args[f]));
				int a, b, c;
				a = -1;
				while ((c = text.read()) != -1)
					{
					b = DigramHist.bucketOf (c);
					if (a != -1 && b != -1)
						hist.increment (a, b);
					a = b;
					}
				text.close();
				}

			// Write histogram to frequency file.
			OutStream out = new OutStream (new FileOutputStream (freqfile));
			out.writeObject (hist);
			out.close();
			}

		else
			{
			// Print histogram.
			System.out.printf ("Digram\tCount\tFrequency%n");
			for (int i = 0; i <= 25; ++ i)
				{
				for (int j = 0; j <= 25; ++ j)
					{
					long count = hist.count (i, j);
					if (count > 0)
						System.out.printf ("%c%c\t%d\t%.6g%n",
							DigramHist.letterOf (i),
							DigramHist.letterOf (j),
							count,
							hist.rowFreq (i, j));
					}
				System.out.printf ("%c total\t%d%n",
					DigramHist.letterOf (i),
					hist.rowTotal(i));
				}
			System.out.printf ("Total\t%d%n", hist.total());
			}
		}

// Hidden operations.

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage:");
		System.err.println ("java AnalyzeDigrams -c <freqfile> <textfile> [<textfile>...]");
		System.err.println ("-- Create frequency file from text file(s)");
		System.err.println ("java AnalyzeDigrams -u <freqfile> <textfile> [<textfile>...]");
		System.err.println ("-- Update frequency file from text file(s)");
		System.err.println ("java AnalyzeDigrams -p <freqfile>");
		System.err.println ("-- Print frequency file");
		System.exit (1);
		}

	}
