//******************************************************************************
//
// File:    AnalyzeStateRecovery02.java
// Package: ---
// Unit:    Class AnalyzeStateRecovery02
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.numeric.Function;
import edu.rit.numeric.ListSeries;
import edu.rit.numeric.ListXYSeries;
import edu.rit.numeric.XYSeries;
import edu.rit.numeric.plot.Plot;
import edu.rit.numeric.plot.Strokes;
import edu.rit.numeric.Mathe;
import edu.rit.util.Map;
import java.awt.Color;
import java.io.File;
import java.util.Scanner;

/**
 * Class AnalyzeStateRecovery02 analyzes the output from a series of runs of the
 * {@linkplain StateRecovery02} program.
 * <P>
 * Usage: <TT>java pj2 AnalyzeStateRecovery02 <I>file</I> [<I>file</I>...]</TT>
 * <P>
 * Each <I>file</I> named on the command line contains the output of one
 * {@linkplain StateRecovery02} run. The program plots conflicts versus number
 * of missing key bits on a log-linear plot.
 *
 * @author  Alan Kaminsky
 * @version 30-Mar-2017
 */
public class AnalyzeStateRecovery02
	{

	/**
	 * Main program.
	 */
	public static void main
		(String[] args)
		throws Exception
		{
		if (args.length < 1) usage();

		ListXYSeries data = new ListXYSeries();
		Map<Integer,ListSeries> dataMap = new Map<Integer,ListSeries>();
		int x = 0;
		int minx = Integer.MAX_VALUE;
		int maxx = Integer.MIN_VALUE;
		long y = 0;
		for (int i = 0; i < args.length; ++ i)
			{
			Scanner s = new Scanner (new File (args[i]));
			while (s.hasNextLine())
				{
				String line = s.nextLine();
				if (line.contains ("missing"))
					{
					x = new Scanner (line) .nextInt();
					}
				else if (line.contains ("conflicts"))
					{
					y = new Scanner (line) .nextLong();
					if (y > 0)
						{
						minx = Math.min (minx, x);
						maxx = Math.max (maxx, x);
						double log2_y = Mathe.log2 (y);
						data.add (x, log2_y);
						ListSeries data_x = dataMap.get (x);
						if (data_x == null)
							{
							data_x = new ListSeries();
							dataMap.put (x, data_x);
							}
						data_x.add (log2_y);
						}
					}
				}
			s.close();
			}

		ListXYSeries medianData = new ListXYSeries();
		for (x = minx; x <= maxx; ++ x)
			{
			ListSeries data_x = dataMap.get (x);
			if (data_x != null)
				medianData.add (x, data_x.robustStats().median);
			}

		Function[] funcs = new Function [4];
		funcs[0] = new Function()
			{
			public double f (double x) { return 1.0; }
			};
		funcs[1] = new Function()
			{
			public double f (double x) { return x; }
			};
		funcs[2] = new Function()
			{
			public double f (double x) { return x*x; }
			};
		funcs[3] = new Function()
			{
			public double f (double x) { return x*x*x; }
			};
		XYSeries.Fit fit = medianData.linearFit (funcs);
		String title = String.format
			("log2(y) = %.5g + %.5g*x + %.5g*x^2 + %.5g*x^3",
			 fit.a[0], fit.a[1], fit.a[2], fit.a[3]);
		System.out.printf ("x = 36, log2(y) = %.5g%n", fit.f(36));

		ListXYSeries fitData = new ListXYSeries();
		for (x = minx; x <= maxx; ++ x)
			fitData.add (x, fit.f(x));

//		Function[] funcs = new Function [2];
//		funcs[0] = new Function()
//			{
//			public double f (double x) { return 1.0; }
//			};
//		funcs[1] = new Function()
//			{
//			public double f (double x)
//				{ return Mathe.logFactorial((long)x)/Math.log(2); }
//			};
//		double[] coeffs = medianData.linearFit (funcs);
//		String title = String.format ("log2(y) = %.5g + %.5g*log2(x!)",
//			coeffs[0], coeffs[1]);
//
//		ListXYSeries fitData = new ListXYSeries();
//		for (x = minx; x <= maxx; ++ x)
//			fitData.add (x, coeffs[0] + coeffs[1]*funcs[1].f(x));

//		int N = medianData.length();
//		ListXYSeries logLogData = new ListXYSeries();
//		for (int i = 0; i < N; ++ i)
//			{
//			double x_i = medianData.x(i);
//			double y_i = medianData.y(i);
//			if (y_i > 0.0)
//				{
//				System.out.printf ("%.5g  %.5g%n",
//					Mathe.log2(medianData.x(i)),
//					Mathe.log2(medianData.y(i)));
//				logLogData.add
//					(Mathe.log2(medianData.x(i)),
//					 Mathe.log2(medianData.y(i)));
//				}
//			}
//		XYSeries.Regression regr = logLogData.linearRegression();
//		double a = Math.pow (2.0, regr.a);
//		double b = regr.b;
//		String title = String.format ("log2(y) = %.5g*x^%.5g", a, b);
//
//		ListXYSeries fitData = new ListXYSeries();
//		for (x = minx; x <= maxx; ++ x)
//			fitData.add (x, a*Math.pow (x, b));

//		int N = medianData.length();
//		ListXYSeries regrData = new ListXYSeries();
//		for (int i = 0; i < N; ++ i)
//			{
//			double x_i = medianData.x(i);
//			double y_i = medianData.y(i);
//			regrData.add (x_i*Mathe.log2(x_i), y_i);
//			}
//		XYSeries.Regression regr = regrData.linearRegression();
//		String title = String.format ("log2(y) = %.5g + %.5g*x*log2(x)",
//			regr.a, regr.b);
//
//		ListXYSeries fitData = new ListXYSeries();
//		for (x = minx; x <= maxx; ++ x)
//			fitData.add (x, regr.a + regr.b*x*Mathe.log2(x));

//		XYSeries.Regression regr = data.linearRegression();
//		System.out.printf ("log2(y) = %.5g + %.5g x, corr = %.5g%n",
//			regr.a, regr.b, regr.corr);
//		double x1 = 0.0;
//		double y1 = regr.a;
//		double x2 = data.maxX();
//		double y2 = regr.a + regr.b*x2;

		System.out.println (title);
		new Plot()
			.plotTitle (title)
			.xAxisTitle ("Missing key letters")
			.xAxisLength (408)
			.yAxisTitle ("log2 (conflicts)")
			.yAxisLength (252)
			.yAxisStart (0)
			.seriesStroke (null)
			.xySeries (data)
			.seriesDots (null)
			.seriesStroke (Strokes.solid (3))
			.seriesColor (Color.BLUE)
			.xySeries (medianData)
			.seriesDots (null)
			.seriesStroke (Strokes.solid (2))
			.seriesColor (Color.RED)
			.xySeries (fitData)
//			.seriesDots (null)
//			.seriesStroke (Strokes.solid (2))
//			.seriesColor (Color.RED)
//			.xySeries (x1, y1, x2, y2)
			.getFrame()
			.setVisible (true);
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java AnalyzeStateRecovery02 <file> [<file>...]");
		System.exit (1);
		}

	}
