//******************************************************************************
//
// File:    DigramHist.java
// Package: ---
// Unit:    Class DigramHist
//
// This Java source file is copyright (C) 2016 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.io.OutStream;
import edu.rit.io.Streamable;
import java.io.IOException;

/**
 * Class DigramHist provides a histogram of English letter digrams.
 * <P>
 * The file <TT>dfreq.dat</TT> in the ElsieFour Library distribution contains a
 * streamed DigramHist object that is a histogram of English letter digrams from
 * the 25 most popular novels from Project Gutenberg.
 *
 * @author  Alan Kaminsky
 * @version 25-Apr-2016
 */
public class DigramHist
	implements Streamable
	{

// Hidden data members.

	private long[][] count = new long [27] [27];
	private long[] rowTotal = new long [27];
	private long total = 0;

// Exported constructors.

	/**
	 * Construct a new digram histogram object.
	 */
	public DigramHist()
		{
		}

// Exported operations.

	/**
	 * Increment the count for the given digram in this histogram.
	 *
	 * @param  i  Bucket index of first letter (0 .. 25).
	 * @param  j  Bucket index of second letter (0 .. 25).
	 *
	 * @see  #bucketOf(int)
	 */
	public void increment
		(int i,
		 int j)
		{
		++ count[i][j];
		++ rowTotal[i];
		++ total;
		}

	/**
	 * Returns the count for the given digram in this histogram.
	 *
	 * @param  i  Bucket index of first letter (0 .. 25).
	 * @param  j  Bucket index of second letter (0 .. 25).
	 *
	 * @return  Count.
	 *
	 * @see  #bucketOf(int)
	 */
	public long count
		(int i,
		 int j)
		{
		return count[i][j];
		}

	/**
	 * Returns the total count for all digrams with the given first letter in
	 * this histogram.
	 *
	 * @param  i  Bucket index of first letter (0 .. 25).
	 *
	 * @return  Total count.
	 *
	 * @see  #bucketOf(int)
	 */
	public long rowTotal
		(int i)
		{
		return rowTotal[i];
		}

	/**
	 * Returns the total count for all digrams in this histogram.
	 *
	 * @return  Total count.
	 */
	public long total()
		{
		return total;
		}

	/**
	 * Returns the frequency of the given digram among all digrams with the
	 * given first letter in this histogram.
	 *
	 * @param  i  Bucket index of first letter (0 .. 25).
	 * @param  j  Bucket index of second letter (0 .. 25).
	 *
	 * @return  Frequency.
	 *
	 * @see  #bucketOf(int)
	 */
	public double rowFreq
		(int i,
		 int j)
		{
		if (rowTotal(i) == 0L)
			return 0.0;
		else
			return (double)count(i,j)/(double)rowTotal(i);
		}

	/**
	 * Returns the frequency of the given digram among all digrams in this
	 * histogram.
	 *
	 * @param  i  Bucket index of first letter (0 .. 25).
	 * @param  j  Bucket index of second letter (0 .. 25).
	 *
	 * @return  Frequency.
	 *
	 * @see  #bucketOf(int)
	 */
	public double freq
		(int i,
		 int j)
		{
		if (total == 0L)
			return 0.0;
		else
			return (double)count(i,j)/(double)total();
		}

	/**
	 * Convert the given character to a bucket index. Letters A .. Z (case
	 * insensitive) convert to bucket indexes 0 .. 25. All other characters
	 * convert to -1 (invalid).
	 *
	 * @param  c  Character.
	 *
	 * @return  Bucket index (0 .. 25), or -1.
	 */
	public static int bucketOf
		(int c)
		{
		if ('A' <= c && c <= 'Z')
			return c - 'A';
		else if ('a' <= c && c <= 'z')
			return c - 'a';
		else
			return -1;
		}

	/**
	 * Convert the given bucket index to a letter. Bucket indexes 0 .. 25
	 * convert to letters A .. Z. All other bucket indexes convert to _
	 * (underscore).
	 *
	 * @param  i  Bucket index.
	 *
	 * @return  Letter.
	 */
	public static char letterOf
		(int i)
		{
		if (0 <= i && i <= 25)
			return (char)(i + 'A');
		else
			return '_';
		}

	/**
	 * Write this histogram to the given out stream.
	 *
	 * @param  out  Out stream.
	 *
	 * @exception  IOException
	 *     Thrown if an I/O error occurred.
	 */
	public void writeOut
		(OutStream out)
		throws IOException
		{
		for (int i = 0; i <= 26; ++ i)
			out.writeLongArray (count[i]);
		out.writeLongArray (rowTotal);
		out.writeLong (total);
		}

	/**
	 * Read this histogram from the given in stream.
	 *
	 * @param  in  In stream.
	 *
	 * @exception  IOException
	 *     Thrown if an I/O error occurred.
	 */
	public void readIn
		(InStream in)
		throws IOException
		{
		for (int i = 0; i <= 26; ++ i)
			in.readLongArray (count[i]);
		in.readLongArray (rowTotal);
		total = in.readLong();
		}

	}
