//******************************************************************************
//
// File:    DigramPicker.java
// Package: ---
// Unit:    Class DigramPicker
//
// This Java source file is copyright (C) 2016 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.util.Random;
import edu.rit.util.Searching;

import edu.rit.io.InStream;
import java.io.File;
import java.io.FileInputStream;

/**
 * Class DigramPicker provides an object for picking random letters according to
 * the distribution in a digram histogram.
 *
 * @author  Alan Kaminsky
 * @version 25-Apr-2016
 */
public class DigramPicker
	{

// Hidden data members.

	private Random prng;
	private double[][] cumul = new double [26] [26];
	private double[] firstCumul = new double [26];
	private Searching.Double helper = new Searching.Double();
	private int prev = -1;

// Exported constructors.

	/**
	 * Construct a new digram picker object. Random numbers are obtained from
	 * the given PRNG. The digram distribution is obtained from the given
	 * histogram.
	 *
	 * @param  prng  Pseudorandom number generator.
	 * @param  hist  Digram histogram object.
	 */
	public DigramPicker
		(Random prng,
		 DigramHist hist)
		{
		this.prng = prng;

		for (int i = 0; i <= 25; ++ i)
			{
			long[] count = new long [26];
			count[0] = hist.count (i, 0);
			for (int j = 1; j <= 25; ++ j)
				count[j] = count[j-1] + hist.count (i, j);
			double total = hist.rowTotal (i);
			for (int j = 0; j <= 25; ++ j)
				cumul[i][j] = (double)count[j]/total;
			}

		long letterTotal = hist.rowTotal (0);
		firstCumul[0] = (double)letterTotal/(double)hist.total();
		for (int i = 1; i <= 25; ++ i)
			{
			letterTotal += hist.rowTotal (i);
			firstCumul[i] = (double)letterTotal/(double)hist.total();
			}
		}

// Exported operations.

	/**
	 * Start a new sequence of randomly chosen letters.
	 */
	public void start()
		{
		prev = -1;
		}

	/**
	 * Returns the next randomly chosen letter as a character A through Z.
	 *
	 * @return  Random character.
	 */
	public char nextChar()
		{
		return DigramHist.letterOf (nextInt());
		}

	/**
	 * Returns the next randomly chosen letter as an integer 0 through 25.
	 *
	 * @return  Random integer.
	 */
	public int nextInt()
		{
		int curr;
		if (prev == -1)
			curr = Searching.searchInterval
				(firstCumul, prng.nextDouble(), helper);
		else
			curr = Searching.searchInterval
				(cumul[prev], prng.nextDouble(), helper);
		prev = curr;
		return curr;
		}

// Unit test main program.

	/**
	 * Unit test main program for class DigramPicker. It reads a streamed digram
	 * histogram object (class {@linkplain DigramHist}) from the given file.
	 * Then it prints <I>N</I> random letters chosen according to the
	 * distribution in the digram histogram.
	 * <P>
	 * Usage: <TT>java DigramPicker <I>file</I> <I>N</I> <I>seed</I></TT>
	 */
	 public static void main
	 	(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		if (args.length != 3) usage();
		File file = new File (args[0]);
		int N = Integer.parseInt (args[1]);
		long seed = Long.parseLong (args[2]);

		// Read histogram from file.
		InStream in = new InStream (new FileInputStream (file));
		DigramHist hist = (DigramHist) in.readObject();
		in.close();

		// Generate random letters.
		Random prng = new Random (seed);
		DigramPicker picker = new DigramPicker (prng, hist);
		for (int i = 0; i < N; ++ i)
			System.out.printf ("%c", picker.nextChar());
		System.out.println();
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java DigramPicker <file> <N> <seed>");
		System.exit (1);
		}

	}
