//******************************************************************************
//
// File:    StatTest02.java
// Package: ---
// Unit:    Class StatTest02
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.io.Tabular;
import edu.rit.numeric.Histogram;
import edu.rit.pj2.Loop;
import edu.rit.pj2.Task;
import edu.rit.util.Random;
import edu.rit.util.RandomSubset;
import java.io.File;
import java.io.FileInputStream;

/**
 * Class StatTest02 is a multicore parallel main program that performs a
 * statistical test on the LC4 cipher.
 * <P>
 * Usage: <TT>java pj2 StatTest02 [-v] <I>dfreq</I> <I>Nlen</I> <I>Plen</I>
 * <I>trials</I> <I>reps</I> <I>seed</I></TT>
 * <P>
 * The program performs <I>reps</I> repetitions in parallel. For each
 * repetition, the program picks a random key. The program does <I>trials</I>
 * trials. For each trial, the program picks a random plaintext of length
 * <I>Plen</I> using the digram frequencies in the <I>dfreq</I> file; picks a
 * random nonce of length <I>Nlen</I>; computes the ciphertext from the key,
 * nonce, and plaintext; and tests whether the ciphertext letters and digrams
 * obey a uniform distribution.
 * <P>
 * For each repetition, the program prints the log Bayes factor of the odds
 * ratio uniformity tests, aggregated across the trials, for the ciphertext
 * letters and digrams. If the <TT>-v</TT> flag is specified, the program also
 * prints the log Bayes factors for each trial individually.
 *
 * @author  Alan Kaminsky
 * @version 24-Mar-2017
 */
public class StatTest02
	extends Task
	{

	// Command line arguments.
	boolean verbose;
	File dfreq;
	int Nlen;
	int Plen;
	int trials;
	int reps;
	long seed;

	// Global variables.
	DigramHist generatorHist;

	// Log Bayes factors for each repetition.
	double[] letterLbf;
	double[] digramLbf;

	/**
	 * Main program.
	 */
	public void main
		(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		int argi = 0;
		while (argi < args.length && args[argi].charAt(0) == '-')
			{
			switch (args[argi])
				{
				case "-v":
					verbose = true;
					break;
				default:
					usage();
				}
			++ argi;
			}
		if (args.length - argi != 6) usage();
		dfreq = new File (args[argi]);
		Nlen = Integer.parseInt (args[argi+1]);
		Plen = Integer.parseInt (args[argi+2]);
		trials = Integer.parseInt (args[argi+3]);
		reps = Integer.parseInt (args[argi+4]);
		seed = Long.parseLong (args[argi+5]);

		// Print provenance.
		System.out.printf ("$ java pj2 StatTest02");
		for (String arg : args) System.out.printf (" %s", arg);
		System.out.println();
		System.out.printf ("Nonce length = %d%n", Nlen);
		System.out.printf ("Plaintext length = %d%n", Plen);
		System.out.printf ("Trials per repetition = %d%n", trials);
		System.out.printf ("Repetitions = %d%n", reps);
		System.out.flush();

		// Read digram probabilities from file.
		InStream in = new InStream (new FileInputStream (dfreq));
		generatorHist = (DigramHist) in.readObject();
		in.close();

		// Set up storage for log Bayes factors.
		letterLbf = new double [reps];
		digramLbf = new double [reps];

		// Do repetitions in parallel.
		parallelFor (0, reps - 1) .schedule (dynamic) .exec (new Loop()
			{
			// Per-thread variables.
			int[] key;
			int[] nonce;
			int[] plaintext;
			int[] ciphertext;
			Histogram letterHist;
			Histogram digramHist;
			LC4 cipher;
			Random prng;
			RandomSubset rs;
			DigramPicker dp;
			double leLbf;
			double diLbf;

			// Initialize per-thread variables.
			public void start()
				{
				// Set up key, nonce, plaintext, ciphertext.
				key = new int [36];
				nonce = new int [Nlen];
				plaintext = new int [Plen];
				ciphertext = new int [Plen];

				// Set up histograms.
				letterHist = new Histogram (36);
				digramHist = new Histogram (1296);

				// Set up cipher.
				cipher = new LC4();

				// Set up pseudorandom number generator.
				prng = new Random (seed + rank());
				rs = new RandomSubset (prng, 36, true);
				dp = new DigramPicker (prng, generatorHist);
				}

			// Perform one repetition.
			public void run (int rep)
				{
				// Initialize log Bayes factors.
				leLbf = 0.0;
				diLbf = 0.0;

				// Generate random key.
				Util.genKey (key, rs);

				// Do trials for random nonces and random plaintexts.
				for (int t = 0; t < trials; ++ t)
					doTrial (rep, t);

				// Record log Bayes factors.
				letterLbf[rep] = leLbf;
				digramLbf[rep] = diLbf;
				}

			// Do one trial.
			private void doTrial
				(int rep,
				 int t)
				{
				// Clear histograms.
				letterHist.clear();
				digramHist.clear();

				// Generate random nonce and random plaintext.
				Util.genNonce (nonce, prng);
				Util.genPlaintext (plaintext, dp);

				// Compute ciphertext.
				cipher.encrypt (key, nonce, plaintext, ciphertext);

				// Accumulate histograms of ciphertext letters.
				for (int i = 0; i < Plen; ++ i)
					{
					int c = ciphertext[i];
					letterHist.accumulate (c);
					}

				// Accumulate histograms of ciphertext digrams.
				for (int i = 0; i < Plen - 1; ++ i)
					{
					int d = ciphertext[i]*36 + ciphertext[i+1];
					digramHist.accumulate (d);
					}

				// Compute log Bayes factors.
				double lbf1 = letterHist.logBayesFactor();
				double lbf2 = digramHist.logBayesFactor();
				leLbf += lbf1;
				diLbf += lbf2;

				// Verbose printout.
				if (verbose)
					{
					System.out.printf ("R%d T%d %.5g %.5g%n",
						rep, t, lbf1, lbf2);
					System.out.flush();
					}
				}
			});

		// Print log Bayes factors for each repetition.
		printPerRepLbfs();
		}

	/**
	 * Print per-repetition log Bayes factors.
	 */
	private void printPerRepLbfs()
		{
		int r;
		double minLetterLbf = Double.POSITIVE_INFINITY;
		double maxLetterLbf = Double.NEGATIVE_INFINITY;
		double minDigramLbf = Double.POSITIVE_INFINITY;
		double maxDigramLbf = Double.NEGATIVE_INFINITY;
		Tabular table = new Tabular();
		table.addColumns (3);
		r = table.addRow();
		table.set (r, 0, "Rep");
		table.set (r, 1, "Letters");
		table.set (r, 2, "Digrams");
		for (int i = 0; i < reps; ++ i)
			{
			r = table.addRow();
			table.setf (r, 0, "%d", i);
			table.setf (r, 1, "%.5g", letterLbf[i]);
			table.setf (r, 2, "%.5g", digramLbf[i]);
			minLetterLbf = Math.min (minLetterLbf, letterLbf[i]);
			maxLetterLbf = Math.max (maxLetterLbf, letterLbf[i]);
			minDigramLbf = Math.min (minDigramLbf, digramLbf[i]);
			maxDigramLbf = Math.max (maxDigramLbf, digramLbf[i]);
			}
		r = table.addRow();
		table.set (r, 0, "Min");
		table.setf (r, 1, "%.5g", minLetterLbf);
		table.setf (r, 2, "%.5g", minDigramLbf);
		r = table.addRow();
		table.set (r, 0, "Max");
		table.setf (r, 1, "%.5g", maxLetterLbf);
		table.setf (r, 2, "%.5g", maxDigramLbf);
		System.out.printf
			("======== Ciphertext log Bayes factors ========%n");
		table.print();
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java pj2 StatTest02 [-v] <dfreq> <Nlen> <Plen> <trials> <reps> <seed>");
		System.err.println ("-v = Print verbose output");
		System.err.println ("<dfreq> = File of digram frequencies");
		System.err.println ("<Nlen> = Nonce length");
		System.err.println ("<Plen> = Plaintext length");
		System.err.println ("<trials> = Number of trials per repetition");
		System.err.println ("<reps> = Number of repetitions");
		System.err.println ("<seed> = Random seed");
		System.exit (1);
		}

	}
